@@ -1,8 +1,8 @@
 #
-# Author:: Adam Jacob (<adam@opscode.com>)
-# Author:: Nuo Yan (<nuo@opscode.com>)
-# Author:: Christopher Brown (<cb@opscode.com>)
-# Copyright:: Copyright (c) 2008 Opscode, Inc.
+# Author:: Adam Jacob (<adam@chef.io>)
+# Author:: Nuo Yan (<nuo@chef.io>)
+# Author:: Christopher Brown (<cb@chef.io>)
+# Copyright:: Copyright (c) 2008-2016 Chef Software, Inc.
 # License:: Apache License, Version 2.0
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -18,11 +18,11 @@
 # limitations under the License.
 #
 
-require 'mixlib/config/version'
-require 'mixlib/config/configurable'
-require 'mixlib/config/unknown_config_option_error'
-require 'mixlib/config/reopened_config_context_with_configurable_error'
-require 'mixlib/config/reopened_configurable_with_config_context_error'
+require "mixlib/config/version"
+require "mixlib/config/configurable"
+require "mixlib/config/unknown_config_option_error"
+require "mixlib/config/reopened_config_context_with_configurable_error"
+require "mixlib/config/reopened_configurable_with_config_context_error"
 
 module Mixlib
   module Config
@@ -30,10 +30,19 @@ module Mixlib
       class << base; attr_accessor :configuration; end
       class << base; attr_accessor :configurables; end
       class << base; attr_accessor :config_contexts; end
+      class << base; attr_accessor :config_context_lists; end
+      class << base; attr_accessor :config_context_hashes; end
       class << base; attr_accessor :config_parent; end
       base.configuration = Hash.new
       base.configurables = Hash.new
       base.config_contexts = Hash.new
+      base.config_context_lists = Hash.new
+      base.config_context_hashes = Hash.new
+      base.initialize_mixlib_config
+    end
+
+    def initialize_mixlib_config
+      @config_strict_mode = nil
     end
 
     # Loads a given ruby file, and runs instance_eval against it in the context of the current
@@ -44,7 +53,58 @@ module Mixlib
     # === Parameters
     # filename<String>:: A filename to read from
     def from_file(filename)
-      self.instance_eval(IO.read(filename), filename, 1)
+      if %w{ .yml .yaml }.include?(File.extname(filename))
+        from_yaml(filename)
+      elsif File.extname(filename) == ".json"
+        from_json(filename)
+      elsif File.extname(filename) == ".toml"
+        from_toml(filename)
+      else
+        instance_eval(IO.read(filename), filename, 1)
+      end
+    end
+
+    # Parses valid YAML structure into Ruby so it can be ingested into the Class
+    #
+    # === Parameters
+    # filename<String>:: A filename to read from
+    def from_yaml(filename)
+      require "yaml"
+      from_hash(YAML.load(IO.read(filename)), filename)
+    end
+
+    # Parses valid JSON structure into Ruby
+    #
+    # === Parameters
+    # filename<String>:: A filename to read from
+    def from_json(filename)
+      require "json"
+      from_hash(JSON.parse(IO.read(filename)), filename)
+    end
+
+    def from_toml(filename)
+      require "tomlrb"
+      from_hash(Tomlrb.parse(IO.read(filename), symbolize_keys: true))
+    end
+
+    # Transforms a Hash into method-style configuration syntax to be processed
+    #
+    # === Parameters
+    # hash<Hash>:: A Hash containing configuration
+    def from_hash(hash, filename = "in_memory")
+      ruby_translation = []
+
+      to_dotted_hash(hash).each do |k, v|
+        if v.is_a? Array
+          ruby_translation << "#{k} #{v}"
+        elsif v.is_a? String
+          ruby_translation << "#{k} \"#{v}\""
+        else
+          ruby_translation << "#{k} #{v}"
+        end
+      end
+
+      instance_eval(ruby_translation.join("\n"), filename, 1)
     end
 
     # Pass Mixlib::Config.configure() a block, and it will yield itself
@@ -52,7 +112,7 @@ module Mixlib
     # === Parameters
     # block<Block>:: A block that is called with self.configuration as the argument.
     def configure(&block)
-      block.call(self.configuration)
+      yield(configuration)
     end
 
     # Get the value of a config option
@@ -93,7 +153,7 @@ module Mixlib
     # <True>:: If the config option exists
     # <False>:: If the config option does not exist
     def has_key?(key)
-      self.configuration.has_key?(key.to_sym)
+      configuration.has_key?(key.to_sym)
     end
 
     # Resets a config option to its default.
@@ -101,13 +161,13 @@ module Mixlib
     # === Parameters
     # symbol<Symbol>:: Name of the config option
     def delete(symbol)
-      self.configuration.delete(symbol)
+      configuration.delete(symbol)
     end
 
     # Resets all config options to their defaults.
     def reset
       self.configuration = Hash.new
-      self.config_contexts.values.each { |config_context| config_context.reset }
+      config_contexts.values.each { |config_context| config_context.reset }
     end
 
     # Makes a copy of any non-default values.
@@ -154,34 +214,71 @@ module Mixlib
     #     }
     #
     def save(include_defaults = false)
-      result = self.configuration.dup
+      result = configuration.dup
       if include_defaults
-        (self.configurables.keys - result.keys).each do |missing_default|
+        (configurables.keys - result.keys).each do |missing_default|
           # Ask any configurables to save themselves into the result array
-          if self.configurables[missing_default].has_default
-            result[missing_default] = self.configurables[missing_default].default
+          if configurables[missing_default].has_default
+            result[missing_default] = configurables[missing_default].default
           end
         end
       end
-      self.config_contexts.each_pair do |key, context|
+      config_contexts.each_pair do |key, context|
         context_result = context.save(include_defaults)
         result[key] = context_result if context_result.size != 0 || include_defaults
       end
+      config_context_lists.each_pair do |key, meta|
+        meta[:values].each do |context|
+          context_result = context.save(include_defaults)
+          result[key] = (result[key] || []) << context_result if context_result.size != 0 || include_defaults
+        end
+      end
+      config_context_hashes.each_pair do |key, meta|
+        meta[:values].each_pair do |context_key, context|
+          context_result = context.save(include_defaults)
+          (result[key] ||= {})[context_key] = context_result if context_result.size != 0 || include_defaults
+        end
+      end
       result
     end
+    alias :to_hash :save
 
     # Restore non-default values from the given hash.
     #
-    # This method is the equivalent of +reset+ followed by +merge!(hash)+.
-    #
     # === Parameters
     # hash<Hash>: a hash in the same format as output by save.
     #
     # === Returns
     # self
     def restore(hash)
-      reset
-      merge!(hash)
+      self.configuration = hash.reject { |key, value| config_contexts.has_key?(key) }
+      config_contexts.each do |key, config_context|
+        if hash.has_key?(key)
+          config_context.restore(hash[key])
+        else
+          config_context.reset
+        end
+      end
+      config_context_lists.each do |key, meta|
+        meta[:values] = []
+        if hash.has_key?(key)
+          hash[key].each do |val|
+            context = define_context(meta[:definition_blocks])
+            context.restore(val)
+            meta[:values] << context
+          end
+        end
+      end
+      config_context_hashes.each do |key, meta|
+        meta[:values] = {}
+        if hash.has_key?(key)
+          hash[key].each do |vkey, val|
+            context = define_context(meta[:definition_blocks])
+            context.restore(val)
+            meta[:values][vkey] = context
+          end
+        end
+      end
     end
 
     # Merge an incoming hash with our config options
@@ -193,11 +290,11 @@ module Mixlib
     # self
     def merge!(hash)
       hash.each do |key, value|
-        if self.config_contexts.has_key?(key)
+        if config_contexts.has_key?(key)
           # Grab the config context and let internal_get cache it if so desired
-          self.config_contexts[key].restore(value)
+          config_contexts[key].restore(value)
         else
-          self.configuration[key] = value
+          configuration[key] = value
         end
       end
       self
@@ -211,7 +308,7 @@ module Mixlib
     # === Returns
     # result of Hash#keys
     def keys
-      self.configuration.keys
+      configuration.keys
     end
 
     # Creates a shallow copy of the internal hash
@@ -280,7 +377,7 @@ module Mixlib
         define_attr_accessor_methods(symbol)
       end
       if block
-        block.call(configurables[symbol])
+        yield(configurables[symbol])
       end
       configurables[symbol]
     end
@@ -322,6 +419,70 @@ module Mixlib
       context
     end
 
+    # Allows you to create a new list of config contexts where you can define new
+    # options with default values.
+    #
+    # This method allows you to open up the configurable more than once.
+    #
+    # For example:
+    #
+    # config_context_list :listeners, :listener do
+    #   configurable(:url).defaults_to("http://localhost")
+    # end
+    #
+    # === Parameters
+    # symbol<Symbol>: the plural name for contexts in the list
+    # symbol<Symbol>: the singular name for contexts in the list
+    # block<Block>: a block that will be run in the context of this new config
+    # class.
+    def config_context_list(plural_symbol, singular_symbol, &block)
+      if configurables.has_key?(symbol)
+        raise ReopenedConfigurableWithConfigContextError, "Cannot redefine config value #{plural_symbol} with a config context"
+      end
+
+      unless config_context_lists.has_key?(plural_symbol)
+        config_context_lists[plural_symbol] = {
+          definition_blocks: [],
+          values: [],
+        }
+        define_list_attr_accessor_methods(plural_symbol, singular_symbol)
+      end
+
+      config_context_lists[plural_symbol][:definition_blocks] << block if block_given?
+    end
+
+    # Allows you to create a new hash of config contexts where you can define new
+    # options with default values.
+    #
+    # This method allows you to open up the configurable more than once.
+    #
+    # For example:
+    #
+    # config_context_hash :listeners, :listener do
+    #   configurable(:url).defaults_to("http://localhost")
+    # end
+    #
+    # === Parameters
+    # symbol<Symbol>: the plural name for contexts in the list
+    # symbol<Symbol>: the singular name for contexts in the list
+    # block<Block>: a block that will be run in the context of this new config
+    # class.
+    def config_context_hash(plural_symbol, singular_symbol, &block)
+      if configurables.has_key?(symbol)
+        raise ReopenedConfigurableWithConfigContextError, "Cannot redefine config value #{plural_symbol} with a config context"
+      end
+
+      unless config_context_hashes.has_key?(plural_symbol)
+        config_context_hashes[plural_symbol] = {
+          definition_blocks: [],
+          values: {},
+        }
+        define_hash_attr_accessor_methods(plural_symbol, singular_symbol)
+      end
+
+      config_context_hashes[plural_symbol][:definition_blocks] << block if block_given?
+    end
+
     NOT_PASSED = Object.new
 
     # Gets or sets strict mode.  When strict mode is on, only values which
@@ -397,17 +558,38 @@ module Mixlib
 
     private
 
+    # Given a (nested) Hash, turn it into a single top-level hash using dots as
+    # nesting notation. This allows for direction translation into method-style
+    # setting of Config.
+    #
+    # === Parameters
+    # hash<Hash>:: The hash to "de-nestify"
+    # recursive_key<String>:: The existing key to prepend going forward
+    #
+    # === Returns
+    # value:: A single-depth Hash using dot notation to indicate nesting
+    def to_dotted_hash(hash, recursive_key = "")
+      hash.each_with_object({}) do |(k , v), ret|
+        key = recursive_key + k.to_s
+        if v.is_a? Hash
+          ret.merge!(to_dotted_hash(v, key + "."))
+        else
+          ret[key] = v
+        end
+      end
+    end
+
     # Internal dispatch setter for config values.
     #
     # === Parameters
     # symbol<Symbol>:: Name of the method (variable setter)
     # value<Object>:: Value to be set in config hash
     #
-    def internal_set(symbol,value)
+    def internal_set(symbol, value)
       if configurables.has_key?(symbol)
-        configurables[symbol].set(self.configuration, value)
+        configurables[symbol].set(configuration, value)
       elsif config_contexts.has_key?(symbol)
-        config_contexts[symbol].restore(value)
+        config_contexts[symbol].restore(value.to_hash)
       else
         if config_strict_mode == :warn
           Chef::Log.warn("Setting unsupported config value #{symbol}.")
@@ -420,9 +602,13 @@ module Mixlib
 
     def internal_get(symbol)
       if configurables.has_key?(symbol)
-        configurables[symbol].get(self.configuration)
+        configurables[symbol].get(configuration)
       elsif config_contexts.has_key?(symbol)
         config_contexts[symbol]
+      elsif config_context_lists.has_key?(symbol)
+        config_context_lists[symbol]
+      elsif config_context_hashes.has_key?(symbol)
+        config_context_hashes[symbol]
       else
         if config_strict_mode == :warn
           Chef::Log.warn("Reading unsupported config value #{symbol}.")
@@ -433,7 +619,7 @@ module Mixlib
       end
     end
 
-    def internal_get_or_set(symbol,*args)
+    def internal_get_or_set(symbol, *args)
       num_args = args.length
       # Setting
       if num_args > 0
@@ -448,13 +634,80 @@ module Mixlib
       # When Ruby 1.8.7 is no longer supported, this stuff can be done with define_singleton_method!
       meta = class << self; self; end
       # Setter
-      meta.send :define_method, "#{symbol.to_s}=".to_sym do |value|
+      meta.send :define_method, "#{symbol}=".to_sym do |value|
         internal_set(symbol, value)
       end
       # Getter
-      meta.send :define_method, symbol do |*args|
-        internal_get_or_set(symbol, *args)
+      meta.send :define_method, symbol do |*args, &block|
+        # If a block was given, eval it in the context
+        if block
+          # If the block expects no arguments, then instance_eval
+          if block.arity == 0
+            internal_get(symbol).instance_eval(&block)
+          else # yield to the block
+            block.yield(internal_get(symbol))
+          end
+        else
+          internal_get_or_set(symbol, *args)
+        end
       end
     end
+
+    def define_list_attr_accessor_methods(plural_symbol, singular_symbol)
+      # When Ruby 1.8.7 is no longer supported, this stuff can be done with define_singleton_method!
+      meta = class << self; self; end
+      # Getter for list
+      meta.send :define_method, plural_symbol do
+        internal_get(plural_symbol)[:values]
+      end
+      # Adds a single new context to the list
+      meta.send :define_method, singular_symbol do |&block|
+        context_list_details = internal_get(plural_symbol)
+        new_context = define_context(context_list_details[:definition_blocks])
+        context_list_details[:values] << new_context
+        # If the block expects no arguments, then instance_eval
+        if block.arity == 0
+          new_context.instance_eval(&block)
+        else # yield to the block
+          block.yield(new_context)
+        end
+      end
+    end
+
+    def define_hash_attr_accessor_methods(plural_symbol, singular_symbol)
+      # When Ruby 1.8.7 is no longer supported, this stuff can be done with define_singleton_method!
+      meta = class << self; self; end
+      # Getter for list
+      meta.send :define_method, plural_symbol do
+        internal_get(plural_symbol)[:values]
+      end
+      # Adds a single new context to the list
+      meta.send :define_method, singular_symbol do |key, &block|
+        context_hash_details = internal_get(plural_symbol)
+        context = if context_hash_details[:values].has_key? key
+                    context_hash_details[:values][key]
+                  else
+                    new_context = define_context(context_hash_details[:definition_blocks])
+                    context_hash_details[:values][key] = new_context
+                    new_context
+                  end
+        # If the block expects no arguments, then instance_eval
+        if block.arity == 0
+          context.instance_eval(&block)
+        else # yield to the block
+          block.yield(context)
+        end
+      end
+    end
+
+    def define_context(definition_blocks)
+      context = Class.new
+      context.extend(::Mixlib::Config)
+      context.config_parent = self
+      definition_blocks.each do |block|
+        context.instance_eval(&block)
+      end
+      context
+    end
   end
 end
